'''ZCC Controller Class.'''
from __future__ import annotations

import asyncio
import json
from json.decoder import JSONDecodeError
from pprint import pformat, pprint
import logging
import socket
import sys
import threading
import time
from typing import Dict, List
import uuid

from zcc.constants import LEVEL_BY_VERBOSITY
from zcc.description import ControlPointDescription
from zcc.device import ControlPointDevice
from zcc.errors import ControlPointError
from zcc.socket import ControlPointSocket
from zcc.protocol import ControlPointProtocol


class ControlPoint:
    '''Represents the ZCC controller which connects to individual devices'''

    def __init__(self, description: ControlPointDescription = None,
                 timeout: int = 2, verbosity: int = 0):

        self.logger = logging.getLogger('ControlPoint')
        if verbosity > 2:
            verbosity = 2
        logging.basicConfig(stream=sys.stderr,
                            level=LEVEL_BY_VERBOSITY[verbosity])

        self.host = description.host
        self.port = description.port

        if not (self.host and self.port):
            raise ControlPointError(
                'ControlPoint initialisation failed - must provide at least host and port.')

        self.brand = description.brand
        self.product = description.product
        self.mac = description.mac
        self.available_tcps = description.available_tcps

        self.timeout = timeout

        self.devices: Dict[ControlPointDevice] = {}

        self.actions_received = 0
        self.properties_received = 0
        self.states_received = 0

        self.device_mac = hex(uuid.getnode())[2:]
        self.access_token = None
        self.authorised = False
        self.session_started = False

        self.socket = None

        self.ready = False

        self.loop = asyncio.get_event_loop()
        self.authorised = self.loop.create_future()
        self.session_started = self.loop.create_future()
        self.properties_ready = self.loop.create_future()
        self.actions_ready = self.loop.create_future()
        self.states_ready = self.loop.create_future()

    @property
    def doors(self) -> List[ControlPointDevice]:
        '''Return an array with all doors'''
        return list(filter(lambda device: device.type == 'garagedoor', self.devices.values()))

    @property
    def fans(self) -> List[ControlPointDevice]:
        '''Return an array with all fans'''
        return list(filter(lambda device: device.type == 'fan', self.devices.values()))

    @property
    def lights(self) -> List[ControlPointDevice]:
        '''Return an array with all lights (i.e. switch or dimmer type)'''
        return list(filter(lambda device:
                           device.type == 'light' or
                           device.type == 'switch' or
                           device.type == 'dimmer', self.devices.values()))

    @property
    def outlets(self) -> List[ControlPointDevice]:
        '''Return an array with all outlets'''
        return list(filter(lambda device: device.type == 'outlet', self.devices.values()))

    async def connect(self) -> bool:
        '''Connect to ZCC, build device table and subscribe to updates'''
        self.logger.info('Connecting to ZCC %s:%d', self.host, self.port)

        if not self.socket:
            try:
                self.socket = ControlPointSocket(self.host, self.port)
            except ConnectionRefusedError:
                return False
            except socket.error:
                return False

        self.socket.subscribe(self)

        self.socket.sendall(ControlPointProtocol.authorise(self.device_mac))

        try:
            await asyncio.wait_for(self.authorised, timeout=2)
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Unable to authorise connection.") from error

        self.socket.sendall(ControlPointProtocol.start(
            self.device_mac, self.access_token))

        try:
            await asyncio.wait_for(self.session_started, timeout=2)
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Unable to start session.") from error

        self.socket.sendall(ControlPointProtocol.get('properties'))

        try:
            await asyncio.wait_for(self.properties_ready, timeout=5)
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Didn't receive any properties.") from error

        self.socket.sendall(ControlPointProtocol.get('actions'))

        try:
            await asyncio.wait_for(self.actions_ready, timeout=5)
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Didn't receive any actions.") from error

        self.socket.sendall(ControlPointProtocol.get('states'))

        try:
            await asyncio.wait_for(self.states_ready, timeout=5)
        except asyncio.exceptions.TimeoutError as error:
            raise ControlPointError(
                "Didn't receive any states.") from error

        await asyncio.sleep(self.timeout)

        self.socket.sendall(ControlPointProtocol.subscribe())

        self.ready = True

        self.logger.info('Connected to ZCC %s:%d with %d/%d/%d actions/properties/states',
                         self.host, self.port,
                         self.actions_received, self.states_received, self.properties_received)

    def __del__(self):
        self.disconnect()

    def describe(self) -> str:
        '''Return a string representation of ZCC including devices'''
        header = '+' + '-' * 130 + '+'
        if self.host:
            description = header + '\n'
            description += '| ControlPoint: %12s        %8s %8s %27s devices        %16s:%-6d %s Tcps    |\n' % (
                self.mac if self.mac else "n/a",
                self.product if self.product else 'n/a',
                self.brand if self.brand else 'n/a',
                str(len(self.devices)),
                self.host,
                self.port,
                str(self.available_tcps) if self.available_tcps else "n/a")
            description += header + '\n'
            for key in self.devices:
                description += self.devices[key].describe()
            return description
        else:
            return 'ControlPoint: not found'

    def disconnect(self):
        '''Disconnect from zimi controller'''
        self.socket.sendall(ControlPointProtocol.unsubscribe())
        self.socket.unsubscribe(self)

    def print_description(self):
        '''Print description of the ZCC controller.'''
        print(self.describe())

    def notify(self, notifier):
        '''Receive a notification of an updated object.'''

        response = notifier.get()

        if response.get(ControlPointProtocol.AUTH_APP_FAILED, None):
            self.logger.info('notify() received: %s', pformat(response))
        if response.get(ControlPointProtocol.AUTH_APP_SUCCESS, None):
            self.logger.info('notify() received: %s', pformat(response))
            self.access_token = response[ControlPointProtocol.AUTH_APP_SUCCESS]['accessToken']
            self.authorised.set_result(True)
        if response.get(ControlPointProtocol.START_SESSION_FAILED, None):
            self.logger.info('notify() received: %s', pformat(response))
        if response.get(ControlPointProtocol.START_SESSION_SUCCESS, None):
            self.logger.info('notify() received: %s', pformat(response))
            self.session_started.set_result(True)
        if response.get(ControlPointProtocol.CONTROLPOINT_ACTIONS, None):
            self.__update_devices(response.get(
                ControlPointProtocol.CONTROLPOINT_ACTIONS, None), 'actions')
        if response.get(ControlPointProtocol.CONTROLPOINT_PROPERTIES, None):
            self.__update_devices(response.get(
                ControlPointProtocol.CONTROLPOINT_PROPERTIES, None), 'properties')
        if response.get(ControlPointProtocol.CONTROLPOINT_STATES, None):
            self.__update_devices(response.get(
                ControlPointProtocol.CONTROLPOINT_STATES, None), 'states')
        if response.get(ControlPointProtocol.CONTROLPOINT_STATES_EVENTS, None):
            self.__update_devices(response.get(
                ControlPointProtocol.CONTROLPOINT_STATES_EVENTS, None), 'states')

    def set(self, identifier: str, action: str, params: object = None):
        '''Sends an action for a device.'''

        message = ControlPointProtocol.set(identifier, action, params)
        self.socket.sendall(message)
        self.logger.info("__set() sent %s request to %s", action, identifier)
        self.logger.debug("%s", message)

    def __str__(self):
        return pformat(vars(self)) + '\n'

    def __update_devices(self, devices, target):
        '''Update device target with JSON data for all devices.'''

        for device in devices:
            updates_made = False
            identifier = device['id']
            if not self.devices.get(identifier):
                self.devices[identifier] = ControlPointDevice(self, identifier)
            if 'actions' in target and self.devices[identifier].actions != device[target]:
                self.devices[identifier].actions = device[target]
                self.actions_received += 1
                updates_made = True
            if 'properties' in target and self.devices[identifier].properties != device[target]:
                self.devices[identifier].properties = device[target]
                self.properties_received += 1
                updates_made = True
            if 'states' in target and self.devices[identifier].states != device[target]:
                self.devices[identifier].states = device[target]
                self.states_received += 1
                updates_made = True
            if updates_made:
                self.logger.info(
                    '__update_devices() updating %s for %s to:\n%s',
                    target, identifier, device[target])
                self.devices[identifier].notify_observers()

            if self.actions_received > 0 and not self.actions_ready.done():
                self.actions_ready.set_result(True)
            if self.properties_received > 0 and not self.properties_ready.done():
                self.properties_ready.set_result(True)
            if self.states_received > 0 and not self.states_ready.done():
                self.states_ready.set_result(True)
