'''A socket connection to a Zimi Controller.'''

import json
from json.decoder import JSONDecodeError
import logging
import queue
import socket
import sys
import threading
import typing

from zcc.errors import ControlPointError


class ControlPointSocket:
    '''A TCP/IP socket that includes recvall method with timeout'''

    def __init__(self, host: str, port: int, timeout: int = None):
        '''Create new TCP/IP socket to host and port with optional timeout.
        Update objects are made available via a FIFO queue.'''

        self.host = host
        self.port = port

        self._observers = []

        self.listen_queu = queue.Queue()

        self.logger = logging.getLogger('ControlPointSocket')

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.sock and timeout:
            self.sock.settimeout(timeout)

        try:
            self.sock.connect((self.host, self.port))
        except ConnectionRefusedError as error:
            self.logger.error(
                'ControlPointSocket() connection refused %s error from %s:%s',
                error, self.host, str(self.port))
            raise error
        except socket.error as error:
            self.logger.error(
                'ControlPointSocket() socket %s error from %s:%s',
                error, self.host, str(self.port))
            raise error

        self.listen_thread = threading.Thread(
            target=self.__listen)
        self.listen_thread.daemon = True

        self.listen_thread.start()

    def __listen(self):
        '''Worker thread to monitor state change events and update objects.'''

        for line in self.recvline():
            try:
                response = json.loads(line).get('response')
                if response:
                    self.listen_queu.put(response)
                    for obs in self._observers:
                        obs.notify(self)
            except JSONDecodeError as error:
                raise ControlPointError(error) from error

    def close(self) -> None:
        '''Close socket'''
        self.sock.close()
        self.sock = None

    def get(self):
        '''Fetch an item from the queue.'''
        return self.listen_queu.get(block=False)

    def recvline(self, single=False):
        '''Receive data and yeild it line by line.'''
        buffering = True
        sock_file = self.sock.makefile()
        while buffering:
            try:
                line = sock_file.readline()
                if single:
                    buffering = False
            except socket.error as error:
                buffering = False
                self.logger.info("recvline() caught %s", error)
            self.logger.debug('recvline() ->\n%s', line)
            yield line

    def recvall(self, eof: bytes = None) -> str:
        '''Receive all data returned on the socket until timeout or end of file'''
        buff = b''
        chunk = None
        while chunk != b'':
            try:
                chunk = self.sock.recv(4096)
                buff += chunk
            except socket.timeout:
                break
            if eof and eof in buff:
                break
        data = buff.decode()
        self.logger.debug('recvall() ->\n%s', data)
        return data

    def sendall(self, message: str):
        '''Send all bytes of string to socket end-point'''
        self.logger.debug('sendall(%s)', message)
        self.sock.sendall(message.encode())

    def subscribe(self, observer):
        '''Subscribe an observer object for state changes.
        Observer object must include notify(self, observable, *args, **kwargs) method.'''
        self._observers.append(observer)

    def unsubscribe(self, observer):
        '''Unsubscribe an observer object.'''
        self._observers.remove(observer)
